#include	<stdio.h>
#include	<stdlib.h>
#include	<assert.h>
#include	<string.h>
#include	<unistd.h>
#include	<errno.h>

#include	<getopt.h>

#include	<tsdb.h>
#include	"version.h"

mkdirp(char	*path)
{
	char	*copy = strdup(path);
	char	*slash;
	for(slash=strchr(copy, '/');slash;slash=strchr(slash+1, '/'))
	{
		if(slash==copy)continue;
		*slash = 0;
		if(mkdir(copy, 0755) != 0 && errno != EEXIST)
		{
			perror(copy);
			exit(-1);
		}
		*slash = '/';
	}
	if(mkdir(copy, 0755) != 0 && errno != EEXIST)
	{
		perror(copy);
		exit(-1);
	}
	free(copy);
}

int	cp_from(char	*path, char	*targetdir, char	*name, int	oblig)
{
	char	targetfile[10240];
	sprintf(targetfile, "%s/%s", targetdir, name);
	FILE	*sf = fopen(path, "rb");
	if(!sf) { if(!oblig)return -1; perror(path); exit(-1); }
	FILE	*tf = fopen(targetfile, "wb");
	if(!tf) { perror(targetfile); exit(-1); }
	fseek(sf, 0, SEEK_END);
	size_t	len = ftell(sf);
	rewind(sf);
	void	*buffer = malloc(len);
	int rlen = fread(buffer, 1, len, sf);
	if(rlen != len) { perror(path); exit(-1); }
	int wlen = fwrite(buffer, 1, len, tf);
	if(wlen != len) { perror(targetfile); exit(-1); }
	fclose(sf);
	fclose(tf);
	printf("%-5zd bytes	%s\n", len, name);
	return 0;
}

int cp(char	*sourcedir, char	*targetdir, char	*name, int	oblig)
{
	char	sourcefile[10240];
	sprintf(sourcefile, "%s/%s", sourcedir, name);
	return cp_from(sourcefile, targetdir, name, oblig);
}

int cp_or_gz(char	*sourcedir, char	*targetdir, char	*name, int	oblig)
{
	if(cp(sourcedir, targetdir, name, 0))
	{
		// couldn't find without .gz...
		char	zipped[10240];
		sprintf(zipped, "%s.gz", name);
		return cp(sourcedir, targetdir, zipped, oblig);
	}
	else return 0;
}

int	exists(char	*path, char	*name)
{
	char	fullpath[10240];
	sprintf(fullpath, "%s/%s", path, name);
	FILE	*f = fopen(fullpath, "r");
	if(f)fclose(f);
	return f?1:0;
}

int	exists_or_gz(char	*path, char	*name)
{
	char	altname[1024];
	sprintf(altname, "%s.gz", name);
	if(exists(path, name))return 1;
	if(exists(path, altname))return 1;
	return 0;
}

make_blank(char	*path, char	*name)
{
	char	fullpath[10240];
	sprintf(fullpath, "%s/%s", path, name);
	FILE	*f = fopen(fullpath, "w");
	if(!f) { perror(fullpath); exit(-1); }
	fclose(f);
	printf("%-5d bytes	%s\n", 0, name);
}

fill_in_missing_relations(struct tsdb	*t)
{
	int i;
	for(i=0;i<t->nrelations;i++)
		if(!exists_or_gz(t->path, t->relations[i].name))
			make_blank(t->path, t->relations[i].name);
}

copy_skeleton(char	*skeleton, char	*target)
{
	// make space for the target
	mkdirp(target);
	// have to copy relations
	cp(skeleton, target, "relations", 1);
	// have to copy item
	cp_or_gz(skeleton, target, "item", 1);
	// if possible:
	cp_or_gz(skeleton, target, "item-set", 0);
	cp_or_gz(skeleton, target, "fold", 0);
	// the rest
	struct tsdb	*t = load_tsdb(target);
	assert(t);
	fill_in_missing_relations(t);
}

build_profile(char	*relations_fname, char	*input_fname, char	*target)
{
	FILE	*fin = input_fname?fopen(input_fname, "r"):stdin;
	if(!fin)
	{
		perror(input_fname);
		exit(-1);
	}

	// make space for the target
	mkdirp(target);
	// have to copy relations file
	cp_from(relations_fname, target, "relations", 1);
	// create a blank 'item' file
	char	path[10240];
	sprintf(path, "%s/item", target);
	FILE	*fit = fopen(path, "w");
	fflush(fit);

	struct tsdb	*t = load_tsdb(target);
	// generate 'item' relation
	struct relation	*it = get_relation(t, "item");
	int	it_id = get_field(it, "i-id", "integer");
	int	it_input = get_field(it, "i-input", "string");
	char	line[102400];
	int	lnum = 0;
	while(fgets(line, 102399, fin) != NULL)
	{
		char	*end = line + strlen(line)-1;
		while(end >= line && *end=='\n' || *end=='\r')
			*end-- = 0;
		lnum++;
		if(!*line)printf("skipping blank input line %d\n", lnum);
		char	*tuple[it->nfields];
		bzero(tuple, sizeof(tuple));
		char	idstr[64];
		sprintf(idstr, "%d", lnum);
		tuple[it_id] = idstr;
		char *encode_tsdb_field(char*);
		tuple[it_input] = encode_tsdb_field(line);
		int i;
		for(i=0;i<it->nfields;i++)
		{
			if(i)fprintf(fit, "@");
			if(!tuple[i])
			{
				char	*type = it->fields[i].type;
				if(!strcasecmp(type, "integer"))tuple[i] = strdup("-1");
				else if(!strcasecmp(type, "string"))tuple[i] = strdup("");
				else if(!strcasecmp(type, "date"))tuple[i] = strdup("23-6-2013 14:28:24");	// XXX fix me
			}
			fprintf(fit, "%s", tuple[i]?:"");
		}
		fprintf(fit, "\n");
		free(tuple[it_input]);
	}
	if(fin != stdin)fclose(fin);
	fclose(fit);

	fill_in_missing_relations(t);
}

usage(char	*myname, int	status)
{
	fprintf(stderr, "usage: %s [ OPTIONS ] destination\n", myname);
	fprintf(stderr, "  -v   show version\n");
	fprintf(stderr, "  -h   show usage information\n");
	fprintf(stderr, "mkprof -s source  destination\n");
	fprintf(stderr, "   creates the 'destination' profile by copying\n"
					"   the 'item' and 'relations' files from the source\n"
					"   profile/skeleton (as well as item-set and fold,\n"
					"   if present).  empty files are created for all\n"
					"   other defined relations.\n");
	fprintf(stderr, "mkprof -r relations_file  [ -i input.txt ]  destination\n");
	fprintf(stderr, "   creates the 'destination' profile by copying\n"
					"   the 'relations' file, populates an 'item' file\n"
					"   with a record for each line in 'input.txt' if\n"
					"   specified, or each line of <stdin>.  empty files\n"
					"   are created for all other defined relations.\n");
	exit(status);
}

main(int	argc, char	*argv[])
{
	int	ch;
	char	*source = NULL,
			*destination = NULL,
			*relations = NULL,
			*input = NULL;
	while( (ch = getopt(argc, argv, "s:r:i:vh")) != -1)switch(ch)
	{
		case	'v':
			printf("mkprof version " VERSION "\n");
			exit(0);
		case	'h':
			usage(argv[0], 0);
		case	's':
			source = optarg;
			break;
		case	'r':
			relations = optarg;
			break;
		case	'i':
			input = optarg;
			break;
		default:
			usage(argv[0], -1);
	}

	if(source && relations)
		usage(argv[0], -1);
	if(input && !relations)
		usage(argv[0], -1);
	if(!source && !relations)
		usage(argv[0], -1);
	if(optind >= argc)usage(argv[0], -1);

	destination = argv[optind];
	if(source)copy_skeleton(source, destination);
	else build_profile(relations, input, destination);

	return 0;
}
