#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<assert.h>
#include	<getopt.h>
#include	<math.h>
#include	<locale.h>

#include	<ace/hash.h>
#include	<ace/mrs.h>

#include	"tsdb.h"

struct result_set
{
	char			*iid;
	int				nresults;
	struct result	*results;
};

struct result
{
	int	id;
	struct mrs	*mrs;
};

// field numbers
int				item_id;
int				item_input;
int				parse_id;
int				parse_run_id;
int				parse_i_id;
int				parse_readings;
int				parse_treal;
int				parse_others;
int				result_parse_id;
int				result_id;
int				result_derivation;
int				result_mrs;
int				result_surface;
int				result_tree;

int same_mrs(struct mrs	*x, struct mrs	*y)
{
	if(mrs_subsumes(x,y) && mrs_subsumes(y,x))return 1;
	else return 0;
}

compare_mrs_lists(char	*iid, struct result	*rx, int	nx, struct result	*ry, int	ny)
{
	int	i, j;
	char	*yhit = calloc(1, ny);

	// if we could serialize the MRSes in a unique way, we could use a hash to make this very fast.
	// it is possible in principle, I believe, but not easy.
	// do it the dumb way.
	for(i=0;i<nx;i++)
	{
		for(j=0;j<ny;j++)
		{
			if(same_mrs(rx[i].mrs, ry[j].mrs))
				break;
		}
		if(j==ny)
		{
			printf("item %s: LHS has unpaired MRS (result %d): ", iid, rx[i].id);
			print_mrs(rx[i].mrs);
		}
		else yhit[j] = 1;
	}
	for(i=0;i<ny;i++)
		if(!yhit[i])
		{
			printf("item %s: RHS has unpaired MRS (result %d): ", iid, ry[i].id);
			print_mrs(ry[i].mrs);
		}
	free(yhit);
}


usage(char	*prog)
{
	fprintf(stderr, "usage: %s profile1 profile2\n", prog);
	exit(-1);
}

struct result_set	get_results(struct relation	*parsex, struct relation	*resultx, char	*iid)
{
	int		nresults = 0;
	char	**results = NULL;
	struct mrs	**mrses = NULL;

	struct result_set	rs;
	rs.iid = strdup(iid);
	rs.nresults = 0;
	rs.results = NULL;

	void result_visitor(char	**result_tuple)
	{
		rs.results = realloc(rs.results, sizeof(struct result) * (rs.nresults+1));
		struct result	*r = &rs.results[rs.nresults++];
		char	*mrsstr = result_tuple[result_mrs];
		if(mrsstr && *mrsstr)
		{
			char	*s = strdup(mrsstr);
			r->mrs = parse_mrs(s);
			free(s);
			if(!r->mrs)
			{
				fprintf(stderr, "unparseable mrs:\n%s\n", mrsstr);
				exit(-1);
			}
		}
		else
		{
			fprintf(stderr, "no stored MRS for result %s of item %s\n", result_tuple[result_id], iid);
		}
		char	*idstr = result_tuple[result_id];
		r->id = atoi(idstr);
	}

	char	**parse_tuple;
	parse_tuple = tsdb_lookup_relation_with_key(parsex, parse_i_id, iid);
	if(!parse_tuple)
		return rs;
	tsdb_visit_relation_with_key(resultx, result_parse_id, parse_tuple[parse_id], result_visitor);

	int rcmp(const struct result	*x, const struct result	*y) { return x->id - y->id; }
	qsort(rs.results, rs.nresults, sizeof(struct result), (int (*)(const void*,const void*))rcmp);

	return rs;
}

main(int	argc, char	*argv[])
{
	if(argc-1 != 2)
		usage(argv[0]);
	setlocale(LC_ALL, "");

	struct tsdb	*tx = load_tsdb(argv[1+0]);
	struct tsdb	*ty = load_tsdb(argv[1+1]);

	if(!tsdb_compatible(tx, ty))
	{
		fprintf(stderr, "those profiles have different schemas\n");
		exit(-2);
	}

	struct relation	*item = get_relation(tx, "item");
	item_id = get_field(item, "i-id", "integer");
	item_input = get_field(item, "i-input", "string");
	//printf("item i-id is field %d\n", item_id);
	//printf("item i-input is field %d\n", item_input);

	printf("loaded %d items\n", item->ntuples);

	struct relation	*parsex = get_relation(tx, "parse");
	struct relation	*parsey = get_relation(ty, "parse");
	parse_id = get_field(parsex, "parse-id", "integer");
	parse_run_id = get_field(parsex, "run-id", "integer");
	parse_i_id = get_field(parsex, "i-id", "integer");
	parse_readings = get_field(parsex, "readings", "integer");
	parse_treal = get_field(parsex, "treal", "integer");
	parse_others = get_field(parsex, "others", "integer");
	//printf("parse readings is field %d\n", parse_readings);

	struct relation	*resultx = get_relation(tx, "result");
	printf("%d left results    %s\n", resultx->ntuples, argv[optind+0]);
	struct relation	*resulty = get_relation(ty, "result");
	printf("%d right results   %s\n", resulty->ntuples, argv[optind+1]);
	result_parse_id = get_field(resultx, "parse-id", "integer");
	result_id = get_field(resultx, "result-id", "integer");
	result_derivation = get_field(resultx, "derivation", "string");
	result_mrs = get_field(resultx, "mrs", "string");
	result_surface = get_field(resultx, "surface", "string");
	result_tree = get_field(resultx, "tree", "string");
	//printf("result parse-id is field %d\n", result_parse_id);

	clear_slab();
	set_default_type_hierarchy(new_type_hierarchy("mrs", "top"));
	commit_slab();

	int	i;
	for(i=0;i<item->ntuples;i++)
	{
		char	*i_id = item->tuples[i][item_id];
		char	*i_input = item->tuples[i][item_input];

		printf("item %s\n", i_id);
		clear_slab();
		struct result_set	rx = get_results(parsex, resultx, i_id);
		struct result_set	ry = get_results(parsey, resulty, i_id);
		compare_mrs_lists(i_id, rx.results, rx.nresults, ry.results, ry.nresults);
	}
	return 0;
}

// hack to remove dependency on glibc 2.14
asm (".symver memcpy, memcpy@GLIBC_2.2.5");

void *__wrap_memcpy(void *dest, const void *src, size_t n)
{ return memcpy(dest, src, n); }
