#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<assert.h>

#include	"tsdb.h"

int	has_relation(struct tsdb	*profile, char	*relation)
{
	int i;
	for(i=0;i<profile->nrelations;i++)
		if(!strcmp(profile->relations[i].name, relation))return 1;
	return 0;
}

int	has_field(struct relation	*rel, char	*field, char	*type)
{
	int	i;
	for(i=0;i<rel->nfields;i++)
		if(!strcmp(rel->fields[i].name, field) && (!type || !strcmp(rel->fields[i].type, type)))
			return 1;
	return 0;
}

void read_relations(struct tsdb	*profile, char	*profile_path)
{
	char	*path = malloc(strlen(profile_path) + 128);
	sprintf(path, "%s/relations", profile_path);
	FILE	*f = fopen(path, "r");

	if(!f)
	{
		perror(path);
		exit(-1);
	}

	char	buffer[1024];
	struct relation	*relation = NULL;
	while(fgets(buffer, 1023, f) != NULL)
	{
		char	*tok = strtok(buffer, " \n");
		if(!tok || !*tok)continue;
		if(strchr(tok, ':'))
		{
			profile->nrelations++;
			profile->relations = realloc(profile->relations, sizeof(struct relation) * profile->nrelations);
			relation = &profile->relations[profile->nrelations-1];
			bzero(relation, sizeof(*relation));
			relation->name = strdup(tok);
			assert(relation->name != NULL);
			*strchr(relation->name, ':') = 0;
			relation->loaded = 0;
			continue;
		}
		char	*field = strdup(tok);
		assert(field != NULL);

		tok = strtok(NULL, " \n");
		if(!tok || !*tok || tok[0]!=':')
		{
			fprintf(stderr, "%s: in relation `%s': no type given for field `%s'\n", path, relation->name, field);
			exit(-1);
		}
		char	*type = strdup(tok+1);
		assert(type != NULL);

		char	*key = strtok(NULL, " \n");
		int		iskey = 0;
		if(key && !strcmp(key, ":key"))iskey = 1;

		relation->nfields++;
		relation->fields = realloc(relation->fields, sizeof(struct field) * relation->nfields);
		struct field	*F = &relation->fields[relation->nfields-1];
		F->name = field;
		F->type = type;
		F->iskey = iskey;
	}

	free(path);
	fclose(f);
}

struct relation	*get_relation(struct tsdb	*profile, char	*name)
{
	int	i;
	for(i=0;i<profile->nrelations;i++)
		if(!strcmp(profile->relations[i].name, name))
		{
			struct relation	*r = &profile->relations[i];
			if(!r->loaded)
				load_relation_data(profile, r);
			return r;
		}
	fprintf(stderr, "ERROR: profile had no relation `%s'\n", name);
	exit(-1);
}

int				get_field(struct relation	*rel, char	*field, char	*type)
{
	int	i;
	for(i=0;i<rel->nfields;i++)
	{
		if(strcmp(rel->fields[i].name, field))continue;
		if(type && strcmp(rel->fields[i].type, type))
		{
			fprintf(stderr, "ERROR: relation `%s': expected field `%s' to have type `%s', but type was actually `%s'\n", rel->name, field, type, rel->fields[i].type);
			exit(-1);
		}
		return i;
	}
	fprintf(stderr, "ERROR: relation `%s' has no field `%s'\n", rel->name, field);
	exit(-1);
}

int	tsdb_rels_compatible(struct relation	*rx, struct relation	*ry)
{
	assert(!strcmp(rx->name, ry->name));
	if(rx->nfields != ry->nfields)return 0;
	int	i;
	for(i=0;i<rx->nfields;i++)
	{
		if(strcmp(rx->fields[i].name, ry->fields[i].name))
			return 0;
		if(strcmp(rx->fields[i].type, ry->fields[i].type))
			return 0;
		if(rx->fields[i].iskey != ry->fields[i].iskey)
			return 0;
	}
	return 1;
}

int	tsdb_compatible(struct tsdb	*x, struct tsdb	*y)
{
	int	i, j;
	if(x->nrelations != y->nrelations)return 0;
	for(i=0;i<x->nrelations;i++)
	{
		// for each 'x' relation...
		for(j=0;j<y->nrelations;j++)
		{
			if(!strcmp(x->relations[i].name, y->relations[j].name))
			{
				// found it; is it compatible?
				if(!tsdb_rels_compatible(&x->relations[i], &y->relations[j]))
					return 0;
				break;
			}
		}
		if(j == y->nrelations)
			// couldn't find it in 'y'.
			return 0;
	}
	return 1;
}
